///
///	@file td.c	@brief tower defense functions
///
///	Copyright (c) 2009 - 2011 by Lutz Sammer.  All Rights Reserved.
///
///	Contributor(s):
///
///	License: AGPLv3
///
///	This program is free software: you can redistribute it and/or modify
///	it under the terms of the GNU Affero General Public License as
///	published by the Free Software Foundation, either version 3 of the
///	License.
///
///	This program is distributed in the hope that it will be useful,
///	but WITHOUT ANY WARRANTY; without even the implied warranty of
///	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
///	GNU Affero General Public License for more details.
///
///	$Id: $
//////////////////////////////////////////////////////////////////////////////

///
///	@defgroup td The tower defense module.
///
///	This module handles playing tower defense.  Didn't belongs into uwm,
///	but I needed a place to play around.
///
/// @{

#include <xcb/xcb.h>
#include "uwm.h"

#ifdef USE_TD				// {

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xcb/xcb_icccm.h>
#include <xcb/xcb_atom.h>
#include <xcb/xcb_aux.h>
#include <xcb/xcb_pixel.h>
#include <xcb/xcb_image.h>

#include "queue.h"
#include "core-array/core-array.h"
#include "core-rc/core-rc.h"

#include "misc.h"
#include "client.h"

#include "pointer.h"
#include "draw.h"
#include "image.h"
#include "icon.h"

#include "td.h"

// ------------------------------------------------------------------------ //

/**
**	|...b...|
**	  ____	  -
**	 /    \	  .
**	/      \  a
**	\\	/  .
**	 \\____/	  _
**
**	a = 2 * r
**	b = s + 2 * h
**	h = sin(30/) * s
**	r = cos(30/) * s
**
**	a = 32 pixel
**	r = a / 2
**	s = r / cos(30/)	16
**	h = sin(30/) * (a / 2) / cos(30/) 
**	h = 9.23
**
**	bc: s(4*a(1) * 30 / 180) * r / c(4*a(1) * 30 / 180)
*/

#define HEX_D	(2  * HEX_R)		///< hex diameter

#define HEX_A	(2 * HEX_R)		///< hex a
#define HEX_B	(2 * HEX_H + HEX_S)	///< hex b
#define HEX_H	14			///< hex h
#define HEX_S	28			///< hex s
#define HEX_R	24			///< hex radis

#define MAP_W	18			///< map width
#define MAP_H	12			///< map height

#define MAP_OX	(-HEX_H - HEX_S - MAP_H)	///< map offset x
#define MAP_OY	(-HEX_R * 2)		///< map offset y

    /// tile x to map x
#define TILE2MAP_X(x) (MAP_OX + (x) * (HEX_H + HEX_S))
    /// tile y to map y
#define TILE2MAP_Y(x, y) (MAP_OY + (y) * HEX_A + ((x) & 1) * HEX_R)

/**
**	Headings.
*/
enum _headings_
{
    HEADING_NORTH,			///< 0/
    HEADING_NORTH_EAST,			///< 60/
    HEADING_SOUTH_EAST,			///< 120/
    HEADING_SOUTH,			///< 180/
    HEADING_SOUTH_WEST,			///< 240/
    HEADING_NORTH_WEST,			///< 300/
};

#define MAX_UNITS 500			///< maximal supported units on map

/**
**	Tower-defense global variables structure typedef.
*/
typedef struct _td_ Td;

/**
**	Tower-defense global variables structure.
*/
struct _td_
{
    xcb_window_t Window;		///< tower-defense window
    xcb_pixmap_t Pixmap;		///< current pixmap (double buffer)
    xcb_pixmap_t Working;		///< working pixmap (double buffer)
    uint16_t Width;			///< window width
    uint16_t Height;			///< window height
    uint32_t Ticks;			///< last tick

    struct _tile_			///< tile structure
    {
	int Distance;			///< distance to end-point

	int8_t Type;			///< field type
	int8_t Tower;			///< tower on field
	int8_t Counter;			///< field counter
	//int Units;			///< units on field

	uint16_t LinkX;			///< link x
	uint16_t LinkY;			///< link y
    } Tiles[MAP_W * MAP_H];		///< tiles of map

    uint16_t StackX;			///< stack x
    uint16_t StackY;			///< stack y

    struct _unit_			///< unit structure
    {
	uint8_t T;			///< unit type
	uint8_t H;			///< heading / direction
	uint8_t Speed;			///< speed
	uint16_t X;			///< unit tile x
	uint16_t Y;			///< unit tile y
	int16_t OX;			///< unit offset subpixel x / 64
	int16_t OY;			///< unit offset subpixel y / 64

	int16_t Health0;		///< full health
	int16_t Health;			///< health
	int16_t Credits;		///< credits
	int16_t Points;			///< points
	int16_t Lives;			///< lives
    } Units[MAX_UNITS];			///< units on the map
    int UnitN;				///< number of units

    Config *Config;			///< current game config

    uint32_t HealthBarPixel;		///< color of health-bar

    struct _tower_type_			///< tower structure
    {
	int16_t Speed;			///< fire speed
	int16_t Range;			///< fire range
	int16_t Damage;			///< fire damage
	int16_t Cost;			///< building cost
	int16_t Time;			///< building time
    } TowerTypes[20];			///< tower types

    //
    //	Waves
    //
    int Wave;				///< current wave
    int WaveGroup;			///< current wave element
    int WaveCounter;			///< wave element counter
    int NextWave;			///< next wave counter
    int GroupInterval;			///< wave group interval

    //
    //	Player
    //
    int Credits;			///< credits to spend
    int Points;				///< points
    int Lives;				///< lives

    //
    //	Debug
    //
    int FrameTime;			///< time spend to update a frame
};

static Td TdVars[1];			///< tower-defense globals

/// offset to neightbor field
/// @{
static const int TdOddAddX[] = { +0, +1, +1, +0, -1, -1 };
static const int TdOddAddY[] = { -1, +0, +1, +1, +1, +0 };
static const int TdEvenAddX[] = { +0, +1, +1, +0, -1, -1 };
static const int TdEvenAddY[] = { -1, -1, +0, +1, +0, -1 };

/// @}

// ------------------------------------------------------------------------ //

/**
**	Map pixel coordinate to map tile.
*/
static void TdPixel2Tile(int x, int y, int *rx, int *ry)
{
    int tx;
    int ty;
    int ox;
    int oy;

    x -= MAP_OX;
    tx = x / (HEX_H + HEX_S);
    ox = x % (HEX_H + HEX_S);

    y -= MAP_OY;
    ty = y / (2 * HEX_R);
    oy = y % (2 * HEX_R);

    if (tx & 1) {
	/*   
	 ** \
	 **  \___
	 **  /
	 ** /
	 */
	if (oy >= HEX_R) {
	    // lower side
	    if (ox < (2 * HEX_H - oy * HEX_H / HEX_R)) {
		tx--;
	    } else {
	    }
	} else {
	    // upper side
	    if (ox < (oy * HEX_H / HEX_R)) {
		tx--;
	    } else {
		ty--;
	    }
	}
    } else {
	/*   ___
	 **  /
	 ** /
	 ** \
	 **  \___
	 */
	if (ox < (HEX_H - oy * HEX_H / HEX_R)) {
	    // upper field
	    --tx;
	    --ty;
	} else if (ox < (-HEX_H + oy * HEX_H / HEX_R)) {
	    // lower field
	    --tx;
	} else {
	    // middle field
	}
    }
    *rx = tx;
    *ry = ty;
}

// ------------------------------------------------------------------------ //

/**
**	Draw hex
**
**	@param x	top-left x-coordinate of surrounding box
**	@param y	top-left y-coordinate of surrounding box
*/
static void TdDrawHex(int x, int y)
{
    xcb_point_t points[7];

    points[0].x = x;
    points[0].y = y + HEX_R;
    points[1].x = HEX_H;
    points[1].y = -HEX_R;
    points[2].x = HEX_S;
    points[2].y = 0;
    points[3].x = HEX_H;
    points[3].y = HEX_R;
    points[4].x = -HEX_H;
    points[4].y = HEX_R;
    points[5].x = -HEX_S;
    points[5].y = 0;
    points[6].x = -HEX_H;
    points[6].y = -HEX_R;

    xcb_poly_line(Connection, XCB_COORD_MODE_PREVIOUS, TdVars->Working, RootGC,
	7, points);
}

/**
**	Draw tower bar.
**
**	@param x	top-left x-coordinate of surrounding box
**	@param y	top-left y-coordinate of surrounding box
**	@param l	length of health 0-100%
*/
static void TdDrawTowerBar(int x, int y, int l)
{
    int bx;
    int by;
    xcb_rectangle_t rectangle;

    bx = x + (HEX_H * 2 + HEX_S) / 2;
    by = y + (HEX_R * 2) / 2;

    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&Colors.PanelBG.Pixel);

    rectangle.x = bx - 16;
    rectangle.y = by + 14;
    rectangle.width = 32;
    rectangle.height = 4;
    xcb_poly_fill_rectangle(Connection, TdVars->Working, RootGC, 1,
	&rectangle);

    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&TdVars->HealthBarPixel);

    rectangle.x = bx - 15;
    rectangle.y = by + 15;
    rectangle.width = (l * 100) / 333;
    rectangle.height = 2;
    xcb_poly_fill_rectangle(Connection, TdVars->Working, RootGC, 1,
	&rectangle);
}

/**
**	Draw tower.
**
**	@param t	tower type
**	@param x	top-left x-coordinate of surrounding box
**	@param y	top-left y-coordinate of surrounding box
**
**	@todo heading of tower
*/
static void TdDrawTower(int t, int x, int y)
{
    xcb_point_t points[5];
    xcb_arc_t arcs[1];

    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&Colors.PanelDown.Pixel);

    x += 10;
    y += 6;

    //
    //	draw common tower base
    //

    points[0].x = x;
    points[0].y = y;
    points[1].x = 37;
    points[1].y = 0;
    points[2].x = 0;
    points[2].y = 37;
    points[3].x = -37;
    points[3].y = 0;
    points[4].x = 0;
    points[4].y = -37;
    xcb_poly_line(Connection, XCB_COORD_MODE_PREVIOUS, TdVars->Working, RootGC,
	5, points);
    points[0].x = x + 1;
    points[0].y = y + 1;
    points[1].x = 35;
    points[1].y = 0;
    points[2].x = 0;
    points[2].y = 35;
    points[3].x = -35;
    points[3].y = 0;
    points[4].x = 0;
    points[4].y = -35;
    xcb_poly_line(Connection, XCB_COORD_MODE_PREVIOUS, TdVars->Working, RootGC,
	5, points);

    //
    //	draw tower turrets
    //
    switch (t) {
	case 0:
	    arcs[0].x = x + 12;
	    arcs[0].y = y + 4;
	    arcs[0].width = 15;
	    arcs[0].height = 30;
	    arcs[0].angle1 = 90 * 64;
	    arcs[0].angle2 = 180 * 64;
	    xcb_poly_fill_arc(Connection, TdVars->Working, RootGC, 1, arcs);

	    points[0].x = x + 3;
	    points[0].y = y + 37 / 2;
	    points[1].x = 18;
	    points[1].y = 0;
	    points[2].x = 0;
	    points[2].y = 1;
	    points[3].x = -18;
	    points[3].y = 0;
	    xcb_poly_line(Connection, XCB_COORD_MODE_PREVIOUS, TdVars->Working,
		RootGC, 4, points);
	    break;
	case 1:
	    arcs[0].x = x + 12;
	    arcs[0].y = y + 11;
	    arcs[0].width = 15;
	    arcs[0].height = 15;
	    arcs[0].angle1 = 0 * 64;
	    arcs[0].angle2 = 360 * 64;
	    xcb_poly_fill_arc(Connection, TdVars->Working, RootGC, 1, arcs);

	    points[0].x = x + 3;
	    points[0].y = y + 37 / 2;
	    points[1].x = 9;
	    points[1].y = 0;
	    points[2].x = 0;
	    points[2].y = 1;
	    points[3].x = -9;
	    points[3].y = 0;
	    xcb_poly_line(Connection, XCB_COORD_MODE_PREVIOUS, TdVars->Working,
		RootGC, 4, points);
	    break;
    }
}

/**
**	Draw health bar
**
**	@param x	top-left x-coordinate of surrounding box
**	@param y	top-left y-coordinate of surrounding box
**	@param l	length of health 0-100%
*/
static void TdDrawUnitHealth(int x, int y, int l)
{
    int bx;
    int by;
    xcb_rectangle_t rectangle;

    bx = x + (HEX_H * 2 + HEX_S) / 2;
    by = y + (HEX_R * 2) / 2;

    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&Colors.PanelBG.Pixel);

    rectangle.x = bx - 6;
    rectangle.y = by + 9;
    rectangle.width = 12;
    rectangle.height = 4;
    xcb_poly_fill_rectangle(Connection, TdVars->Working, RootGC, 1,
	&rectangle);

    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&TdVars->HealthBarPixel);

    rectangle.x = bx - 5;
    rectangle.y = by + 10;
    rectangle.width = l / 10;
    rectangle.height = 2;
    xcb_poly_fill_rectangle(Connection, TdVars->Working, RootGC, 1,
	&rectangle);
}

/**
**	Draw unit.
*/
static void TdDrawUnit(int t, int x, int y, int h)
{
    xcb_coloritem_t temp;
    xcb_point_t points[5];
    xcb_arc_t arcs[1];
    int bx;
    int by;

    // x' =  x*cos(angle) - y*sin(angle)
    // y' =  x*sin(angle) + y*cos(angle)
    static const int TdSin[] = { 0, 866, 866, 0, -866, -866 };
    static const int TdCos[] = { 1000, 500, -500, -1000, -500, 500 };

    bx = x + (HEX_H * 2 + HEX_S) / 2;
    by = y + (HEX_R * 2) / 2;

    switch (t) {
	case 0:			// circle
	    ColorParse("yellow", &temp);
	    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND, &temp.pixel);

	    arcs[0].x = x + ((HEX_H * 2 + HEX_S) - 18) / 2;
	    arcs[0].y = y + ((HEX_R * 2) - 18) / 2;
	    arcs[0].width = 18;
	    arcs[0].height = 18;
	    arcs[0].angle1 = 0 * 64;
	    arcs[0].angle2 = 360 * 64;
	    xcb_poly_fill_arc(Connection, TdVars->Working, RootGC, 1, arcs);
	    break;
	case 1:			// triangle
	    ColorParse("green", &temp);
	    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND, &temp.pixel);

	    points[0].x = bx + (0 * TdCos[h]) / 1000 + (-8 * TdSin[h]) / 1000;
	    points[0].y = by - (0 * TdSin[h]) / 1000 + (-8 * TdCos[h]) / 1000;
	    points[1].x = bx + (9 * TdCos[h]) / 1000 + (8 * TdSin[h]) / 1000;
	    points[1].y = by - (9 * TdSin[h]) / 1000 + (8 * TdCos[h]) / 1000;
	    points[2].x = bx + (-9 * TdCos[h]) / 1000 + (8 * TdSin[h]) / 1000;
	    points[2].y = by - (-9 * TdSin[h]) / 1000 + (8 * TdCos[h]) / 1000;
	    xcb_fill_poly(Connection, TdVars->Working, RootGC, 0,
		XCB_COORD_MODE_ORIGIN, 3, points);
	    break;
	case 2:			// rectangle
	    ColorParse("orange", &temp);
	    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND, &temp.pixel);

	    points[0].x = bx + (-9 * TdCos[h]) / 1000 + (-9 * TdSin[h]) / 1000;
	    points[0].y = by - (-9 * TdSin[h]) / 1000 + (-9 * TdCos[h]) / 1000;
	    points[1].x = bx + (9 * TdCos[h]) / 1000 + (-9 * TdSin[h]) / 1000;
	    points[1].y = by - (9 * TdSin[h]) / 1000 + (-9 * TdCos[h]) / 1000;
	    points[2].x = bx + (9 * TdCos[h]) / 1000 + (9 * TdSin[h]) / 1000;
	    points[2].y = by - (9 * TdSin[h]) / 1000 + (9 * TdCos[h]) / 1000;
	    points[3].x = bx + (-9 * TdCos[h]) / 1000 + (9 * TdSin[h]) / 1000;
	    points[3].y = by - (-9 * TdSin[h]) / 1000 + (9 * TdCos[h]) / 1000;

	    xcb_fill_poly(Connection, TdVars->Working, RootGC, 0,
		XCB_COORD_MODE_ORIGIN, 4, points);
	    break;
    }
}

/**
**	Draw the hex grid.
*/
static void TdDrawGrid(void)
{
    int x;
    int y;

    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&Colors.PanelFG.Pixel);

    for (y = 0; y < MAP_H; ++y) {
	for (x = 0; x < MAP_W; ++x) {
	    switch (TdVars->Tiles[x + y * MAP_W].Type) {
		case '.':
		case ' ':
		    TdDrawHex(TILE2MAP_X(x), TILE2MAP_Y(x, y));
		default:
		    break;
	    }
	}
    }
}

#if 0

/**
**	Draw debug info.
*/
static void TdDrawDebug(void)
{
    int x;
    int y;

    for (y = 0; y < MAP_H; ++y) {
	for (x = 0; x < MAP_W; ++x) {
	    char buf[32];

	    sprintf(buf, "%d,%d:%d", x, y,
		TdVars->Tiles[x + y * MAP_W].Distance);
	    sprintf(buf, "%d", TdVars->Tiles[x + y * MAP_W].Distance);
	    FontDrawString(TdVars->Working, &Fonts.Pager, Colors.PanelFG.Pixel,
		TILE2MAP_X(x), TILE2MAP_Y(x, y), HEX_B, NULL, buf);
	}
    }
}
#endif

/**
**	Draw cursor.
*/
static void TdDrawCursor(void)
{
    int x;
    int y;
    int tx;
    int ty;

    PointerGetPosition(&x, &y);
    TdPixel2Tile(x, y, &tx, &ty);
    if (tx < 0 || ty < 0 || tx >= MAP_W || ty >= MAP_H) {
	return;
    } else if (TdVars->Tiles[tx + ty * MAP_W].Distance == 9999) {
	return;
    } else {
	int tt;

	//
	//	Draw range of tower
	//
	if ((tt = TdVars->Tiles[tx + ty * MAP_W].Tower)) {
	    xcb_arc_t arcs[1];

	    --tt;
	    x = TILE2MAP_X(tx);
	    y = TILE2MAP_Y(tx, ty);

	    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
		&Colors.PanelUp.Pixel);

	    arcs[0].x =
		x - TdVars->TowerTypes[tt].Range + (HEX_H * 2 + HEX_S) / 2;
	    arcs[0].y = y - TdVars->TowerTypes[tt].Range + (HEX_R * 2) / 2;
	    arcs[0].width = TdVars->TowerTypes[tt].Range * 2;
	    arcs[0].height = TdVars->TowerTypes[tt].Range * 2;
	    arcs[0].angle1 = 0 * 64;
	    arcs[0].angle2 = 360 * 64;
	    xcb_poly_arc(Connection, TdVars->Working, RootGC, 1, arcs);

	    return;
	}
    }
}

/**
**	Draw all units.
*/
static void TdDrawUnits(void)
{
    int i;

    // TdDrawUnit(2, TILE2MAP_X(2), TILE2MAP_Y(2, 2), 0);

    for (i = 0; i < TdVars->UnitN; ++i) {
	int x;
	int y;

	x = TILE2MAP_X(TdVars->Units[i].X) + TdVars->Units[i].OX / 64;
	y = TILE2MAP_Y(TdVars->Units[i].X, TdVars->Units[i].Y)
	    + TdVars->Units[i].OY / 64;
	TdDrawUnit(TdVars->Units[i].T, x, y, TdVars->Units[i].H);
	TdDrawUnitHealth(x, y,
	    (TdVars->Units[i].Health * 100) / TdVars->Units[i].Health0);
    }
}

/**
**	Draw all towers.
*/
static void TdDrawTowers(void)
{
    int x;
    int y;

    for (y = 0; y < MAP_H; ++y) {
	for (x = 0; x < MAP_W; ++x) {
	    int tt;

	    if ((tt = TdVars->Tiles[x + y * MAP_W].Tower)) {
		--tt;
		TdDrawTower(abs(tt), TILE2MAP_X(x), TILE2MAP_Y(x, y));
		if (tt >= 0) {
		    TdDrawTowerBar(TILE2MAP_X(x), TILE2MAP_Y(x, y),
			((TdVars->TowerTypes[tt].Speed - TdVars->Tiles[x +
				    y * MAP_W].Counter) * 100)
			/ TdVars->TowerTypes[tt].Speed);
		}
	    }
	}
    }
}

// ------------------------------------------------------------------------ //

/**
**	Draw the HUD (Heads-Up Display).
*/
static void TdDrawHud(void)
{
    char buf[128];

    sprintf(buf, "Wave: %3d Next in %4d", TdVars->Wave, TdVars->NextWave);
    FontDrawString(TdVars->Working, &Fonts.Panel, Colors.PanelFG.Pixel, 0, 0,
	TdVars->Width / 2, NULL, buf);
    sprintf(buf, "Credits: %8d Lives: %3d", TdVars->Credits, TdVars->Lives);
    FontDrawString(TdVars->Working, &Fonts.Panel, Colors.PanelFG.Pixel, 0,
	TdVars->Height - Fonts.Panel.Height, TdVars->Width / 2, NULL, buf);

    sprintf(buf, "Frame: %3d", TdVars->FrameTime);
    FontDrawString(TdVars->Working, &Fonts.Panel, Colors.PanelFG.Pixel,
	TdVars->Width - 100, 0, TdVars->Width / 2, NULL, buf);
}

/**
**	Draw tower-defense.
**
**	@param expose	flag called from expose event
*/
static void TdDraw(int expose)
{
    uint32_t values[1];
    xcb_rectangle_t rectangle;

    /*
       if (IsNextEventAvail()) {
       return;
       }
     */

    if (expose) {
	return;
    }
    // clear with background
    xcb_change_gc(Connection, RootGC, XCB_GC_FOREGROUND,
	&Colors.PanelBG.Pixel);
    rectangle.x = 0;
    rectangle.y = 0;
    rectangle.width = TdVars->Width;
    rectangle.height = TdVars->Height;
    xcb_poly_fill_rectangle(Connection, TdVars->Working, RootGC, 1,
	&rectangle);

    TdDrawGrid();
    // TdDrawDebug();
    TdDrawUnits();
    TdDrawTowers();
    TdDrawHud();
    TdDrawCursor();

    //	swap pixmaps
    values[0] = TdVars->Working;
    xcb_change_window_attributes(Connection, TdVars->Window,
	XCB_CW_BACK_PIXMAP, values);
    xcb_aux_clear_window(Connection, TdVars->Window);
    TdVars->Working = TdVars->Pixmap;
    TdVars->Pixmap = values[0];
}

// ------------------------------------------------------------------------ //

/**
**	Prepare map.
**
**	@todo FIXME: width, height isn't configurable yet
*/
static void TdPrepareMap(void)
{
    int x;
    int y;
    const ConfigObject *array;
    const ConfigObject *index;
    const ConfigObject *value;
    ssize_t ival;

    if (!ConfigGetArray(ConfigDict(TdVars->Config), &array, "map", NULL)) {
	Warning("no map found\n");
	return;
    }
    if (ConfigGetInteger(array, &ival, "width", NULL)) {
	if (ival != MAP_W) {
	    Warning("map width conflict\n");
	    return;
	}
    }
    if (ConfigGetInteger(array, &ival, "height", NULL)) {
	if (ival != MAP_H) {
	    Warning("map height conflict\n");
	    return;
	}
    }

    index = 0;
    value = ConfigArrayFirstFixedKey(array, &index);

    memset(TdVars->Tiles, 0, sizeof(TdVars->Tiles));
    for (y = 0; y < MAP_H; ++y) {
	const char *sval;

	if (ConfigCheckString(value, &sval)) {
	    if (strlen(sval) != MAP_W) {
		Warning("map string conflict\n");
	    } else
		for (x = 0; x < MAP_W; ++x) {
		    TdVars->Tiles[x + y * MAP_W].Type = sval[x];
		}
	}
	value = ConfigArrayNextFixedKey(array, &index);
    }
}

/**
**	Push coordinates.
**
**	@param x	tile x-coordinate
**	@param y	tile y-coordinate
*/
static inline void TdPush(int x, int y)
{
    TdVars->Tiles[x + y * MAP_W].LinkX = TdVars->StackX;
    TdVars->Tiles[x + y * MAP_W].LinkY = TdVars->StackY;
    TdVars->StackX = x;
    TdVars->StackY = y;
}

/**
**	Pop coordinates.
**
**	@param x	tile x-coordinate result
**	@param y	tile y-coordinate result
*/
static inline void TdPop(int *x, int *y)
{
    int rx;
    int ry;

    rx = TdVars->Tiles[*x + *y * MAP_W].LinkX;
    ry = TdVars->Tiles[*x + *y * MAP_W].LinkY;

    *x = rx;
    *y = ry;
}

/**
**	Try movement.
*/
static inline void TdTry(int x, int y, int d)
{
    // printf("Try %d, %d\n", x, y);
    // out side map
    if (x < 0 || y < 0 || x >= MAP_W || y >= MAP_H) {
	return;
    }
    // already visited
    if (TdVars->Tiles[x + y * MAP_W].Distance) {
	return;
    }
    TdVars->Tiles[x + y * MAP_W].Distance = d;
    TdPush(x, y);
}

/**
**	Movement map.
*/
static void TdMoveMap(void)
{
    int x;
    int y;
    int d;

    TdVars->StackX = 0;
    TdVars->StackY = 0;
    d = 1;

    //
    //	Reset map and mark border fields.
    //
    for (y = 0; y < MAP_H; ++y) {
	for (x = 0; x < MAP_W; ++x) {
	    TdVars->Tiles[x + y * MAP_W].Distance = 0;
	    switch (TdVars->Tiles[x + y * MAP_W].Type) {
		case '.':		// empty field
		    if (TdVars->Tiles[x + y * MAP_W].Tower) {
			TdVars->Tiles[x + y * MAP_W].Distance = 9998;
		    }
		case 'E':		// entry field
		    break;
		case 'L':		// leave field
		    TdVars->Tiles[x + y * MAP_W].Distance = 1;
		    //
		    //	Push end=exit points on stack.
		    //
		    TdPush(x, y);
		    break;
		default:		// border or blocked field
		    TdVars->Tiles[x + y * MAP_W].Distance = 9999;
		    break;
	    }
	}
    }

    // Loop until all tiles are marked
    for (;;) {
	x = TdVars->StackX;
	y = TdVars->StackY;
	if (!x && !y) {
	    break;
	}

	TdVars->StackX = 0;
	TdVars->StackY = 0;
	++d;
	while (x || y) {
	    int h;

	    // printf("Pop %d, %d\n", x, y);
	    // try all 6 surrounded tiles
	    if (x & 1) {
		for (h = 0; h < 6; ++h) {
		    TdTry(x + TdOddAddX[h], y + TdOddAddY[h], d);
		}
	    } else {
		for (h = 0; h < 6; ++h) {
		    TdTry(x + TdEvenAddX[h], y + TdEvenAddY[h], d);
		}
	    }
	    TdPop(&x, &y);
	}
    }
}

// ------------------------------------------------------------------------ //

///	Heading x offsets.
static const int TdHeadingOX[] = {
    0,
    -64 * (HEX_H + HEX_S),
    -64 * (HEX_H + HEX_S),
    0,
    +64 * (HEX_H + HEX_S),
    +64 * (HEX_H + HEX_S),
};

///	Heading y offsets.
static const int TdHeadingOY[] = {
    +64 * HEX_A,
    +64 * HEX_R,
    -64 * HEX_R,
    -64 * HEX_A,
    -64 * HEX_R,
    +64 * HEX_R,
};

/**
**	New unit.
*/
static int TdNewUnit(int t, int x, int y)
{
    int n;

    if ((n = TdVars->UnitN) == MAX_UNITS) {
	return -1;
    }

    TdVars->Units[n].T = t;
    TdVars->Units[n].H = HEADING_NORTH;
    TdVars->Units[n].Speed = 1 * 64;
    TdVars->Units[n].X = x;
    TdVars->Units[n].Y = y;
    TdVars->Units[n].OX = 0;
    TdVars->Units[n].OY = 0;

    TdVars->Units[n].Health0 = 10;
    TdVars->Units[n].Health = 10;
    TdVars->Units[n].Credits = 1;
    TdVars->Units[n].Points = 1;
    TdVars->Units[n].Lives = 1;

    TdVars->UnitN = n + 1;

    return n;
}

/**
**	Del unit.
*/
void TdDelUnit(int i)
{
    if (TdVars->UnitN-- >= 2) {
	TdVars->Units[i] = TdVars->Units[TdVars->UnitN];
    }
}

/**
**	Move & animate units.
*/
void TdMoveUnits(void)
{
    int i;

    for (i = 0; i < TdVars->UnitN; ++i) {
	// reached tile
	if (!TdVars->Units[i].OX && !TdVars->Units[i].OY) {
	    int x;
	    int y;
	    int d;
	    int h;
	    int h0;

	    // search tile to move to.
	    x = TdVars->Units[i].X;
	    y = TdVars->Units[i].Y;
	    d = TdVars->Tiles[x + y * MAP_W].Distance;
	    h0 = rand() % 6;

	    // try all 6 surrounded tiles
	    if (x & 1) {
		for (h = h0; h < h0 + 6; ++h) {
		    x += TdOddAddX[h % 6];
		    y += TdOddAddY[h % 6];
		    // inside map
		    if (x >= 0 && y >= 0 && x < MAP_W && y < MAP_H) {
			if (TdVars->Tiles[x + y * MAP_W].Distance < d) {
			    h %= 6;
			    goto found;
			}
		    }
		    x -= TdOddAddX[h % 6];
		    y -= TdOddAddY[h % 6];
		}
	    } else {
		for (h = h0; h < h0 + 6; ++h) {
		    static const int ix[] = { +0, +1, +1, +0, -1, -1 };
		    static const int iy[] = { -1, -1, +0, +1, +0, -1 };

		    x += ix[h % 6];
		    y += iy[h % 6];
		    // inside map
		    if (x >= 0 && y >= 0 && x < MAP_W && y < MAP_H) {
			if (TdVars->Tiles[x + y * MAP_W].Distance < d) {
			    h %= 6;
			    goto found;
			}
		    }
		    x -= ix[h % 6];
		    y -= iy[h % 6];
		}
	    }
	    //
	    //	reached endpoint
	    //
	    if (TdVars->Tiles[x + y * MAP_W].Distance == 1) {
		//printf("reached endpoint\n");
		// FIXME: Handle counting, move outside the map
		TdDelUnit(i);
		// --i;
		continue;
	    } else if (h == h0 + 6) {
		printf("oops\n");
		h = 0;
	    }
	  found:
	    TdVars->Units[i].X = x;
	    TdVars->Units[i].Y = y;
	    TdVars->Units[i].H = h;
	    TdVars->Units[i].OX = TdHeadingOX[h];
	    TdVars->Units[i].OY = TdHeadingOY[h];
	}

	switch (TdVars->Units[i].H) {
	    case HEADING_NORTH:	// N
		TdVars->Units[i].OY -=
		    (TdVars->Units[i].Speed * HEX_A) / HEX_D;
		if (TdVars->Units[i].OY < 0) {
		    TdVars->Units[i].OY = 0;
		}
		break;
	    case HEADING_NORTH_EAST:	// NE
		TdVars->Units[i].OX +=
		    (TdVars->Units[i].Speed * (HEX_H + HEX_S)) / HEX_D;
		if (TdVars->Units[i].OX > 0) {
		    TdVars->Units[i].OX = 0;
		}
		TdVars->Units[i].OY -=
		    (TdVars->Units[i].Speed * HEX_R) / HEX_D;
		if (TdVars->Units[i].OY < 0) {
		    TdVars->Units[i].OY = 0;
		}
		break;
	    case HEADING_SOUTH_EAST:	// SE
		TdVars->Units[i].OX +=
		    (TdVars->Units[i].Speed * (HEX_H + HEX_S)) / HEX_D;
		if (TdVars->Units[i].OX > 0) {
		    TdVars->Units[i].OX = 0;
		}
		TdVars->Units[i].OY +=
		    (TdVars->Units[i].Speed * HEX_R) / HEX_D;
		if (TdVars->Units[i].OY > 0) {
		    TdVars->Units[i].OY = 0;
		}
		break;
	    case HEADING_SOUTH:	// S
		TdVars->Units[i].OY +=
		    (TdVars->Units[i].Speed * HEX_A) / HEX_D;
		if (TdVars->Units[i].OY > 0) {
		    TdVars->Units[i].OY = 0;
		}
		break;
	    case HEADING_SOUTH_WEST:	// SW
		TdVars->Units[i].OX -=
		    (TdVars->Units[i].Speed * (HEX_H + HEX_S)) / HEX_D;
		if (TdVars->Units[i].OX < 0) {
		    TdVars->Units[i].OX = 0;
		}
		TdVars->Units[i].OY += (TdVars->Units[i].Speed * HEX_R)
		    / HEX_D;
		if (TdVars->Units[i].OY > 0) {
		    TdVars->Units[i].OY = 0;
		}
		break;
		break;
	    case HEADING_NORTH_WEST:	// NW
		TdVars->Units[i].OX -=
		    (TdVars->Units[i].Speed * (HEX_H + HEX_S)) / HEX_D;
		if (TdVars->Units[i].OX < 0) {
		    TdVars->Units[i].OX = 0;
		}
		TdVars->Units[i].OY -=
		    (TdVars->Units[i].Speed * HEX_R) / HEX_D;
		if (TdVars->Units[i].OY < 0) {
		    TdVars->Units[i].OY = 0;
		}
		break;
	}
    }
}

// ------------------------------------------------------------------------ //

/**
**	Update fire/build/sell all towers.
**
**	@todo can use the link field of tile to link all towers.
*/
static void TdUpdateTowers(void)
{
    int x;
    int y;
    int tt;
    int i;

    for (y = 0; y < MAP_H; ++y) {
	for (x = 0; x < MAP_W; ++x) {
	    if ((tt = TdVars->Tiles[x + y * MAP_W].Tower) > 0) {
		if (TdVars->Tiles[x + y * MAP_W].Counter) {
		    TdVars->Tiles[x + y * MAP_W].Counter--;
		} else {
		    // Check if some unit in range.
		    int tower_x;
		    int tower_y;
		    int tower_r;

		    --tt;
		    // Tower-center
		    tower_x = TILE2MAP_X(x);	// + (HEX_H * 2 + HEX_S) / 2;
		    tower_y = TILE2MAP_Y(x, y);	// + (HEX_R * 2) / 2;
		    tower_r = TdVars->TowerTypes[tt].Range;
		    tower_r *= tower_r;

		    //
		    //	Check all units against in range.
		    //
		    for (i = 0; i < TdVars->UnitN; ++i) {
			int unit_x;
			int unit_y;
			int d;

			unit_x =
			    TILE2MAP_X(TdVars->Units[i].X) +
			    TdVars->Units[i].OX / 64;
			unit_y =
			    TILE2MAP_Y(TdVars->Units[i].X, TdVars->Units[i].Y)
			    + TdVars->Units[i].OY / 64;

			d = (unit_x - tower_x) * (unit_x - tower_x)
			    + (unit_y - tower_y) * (unit_y - tower_y);
			if (d < tower_r) {
			    TdVars->Units[i].Health -=
				TdVars->TowerTypes[tt].Damage;
			    if (TdVars->Units[i].Health < 0) {
				printf("Kill %d\n", i);

				TdVars->Points += TdVars->Units[i].Points;
				TdVars->Credits += TdVars->Units[i].Credits;
				TdDelUnit(i);
			    }

			    TdVars->Tiles[x + y * MAP_W].Counter =
				TdVars->TowerTypes[tt].Speed;
			    break;
			}
		    }
		    //printf("Fire\n");
		}
	    }
	}
    }
}

// ------------------------------------------------------------------------ //

/**
**	Create wave unit.
**
**	@param array	config of new wave unit
*/
static void TdCreateUnit(const ConfigObject * array)
{
    int lx;
    int ly;
    int x;
    int y;
    int i;
    int type;
    ssize_t ival;

    //
    //	Find entry point
    //
    lx = 0;
    ly = 0;
    i = 0;
    for (y = 0; y < MAP_H; ++y) {
	for (x = 0; x < MAP_W; ++x) {
	    if (TdVars->Tiles[x + y * MAP_W].Type == 'E') {
		TdVars->Tiles[x + y * MAP_W].LinkX = lx;
		TdVars->Tiles[x + y * MAP_W].LinkY = ly;
		lx = x;
		ly = y;
		++i;
	    }
	}
    }

    if (!i) {				// no entry point
	return;
    }
    //
    //	Choose random entry point? or just circle
    //
    i = rand() % i;

    x = lx;
    y = ly;
    while (i--) {
	lx = TdVars->Tiles[x + y * MAP_W].LinkX;
	ly = TdVars->Tiles[x + y * MAP_W].LinkY;
	x = lx;
	y = ly;
    }

    type = 0;
    if (ConfigGetInteger(array, &ival, "type", NULL)) {
	type = ival;
    }
    i = TdNewUnit(type, x, y);
    if (i >= 0) {
	if (ConfigGetInteger(array, &ival, "speed", NULL)) {
	    TdVars->Units[i].Speed = ival;
	}
	if (ConfigGetInteger(array, &ival, "health", NULL)) {
	    TdVars->Units[i].Health0 = TdVars->Units[i].Health = ival;
	}
	if (ConfigGetInteger(array, &ival, "points", NULL)) {
	    TdVars->Units[i].Points = ival;
	}
	if (ConfigGetInteger(array, &ival, "credits", NULL)) {
	    TdVars->Units[i].Credits = ival;
	}
	if (ConfigGetInteger(array, &ival, "lives", NULL)) {
	    TdVars->Units[i].Lives = ival;
	}
    }
}

/**
**	Handle new waves.
*/
static void TdUpdateWave(void)
{
    const ConfigObject *array;
    const ConfigObject *index;
    const ConfigObject *value;
    ssize_t ival;

    //
    //	Time to start next wave
    //
    if (!TdVars->NextWave) {
	if (!ConfigGetArray(ConfigDict(TdVars->Config), &array, "wave", NULL)) {
	    Warning("no wave found\n");
	    return;
	}
	// waves
	index = ConfigNewInteger(TdVars->Wave);
	array = ConfigArrayFirstFixedKey(array, &index);
	if (!array) {
	    Warning("wave %d not found\n", TdVars->Wave);
	    // FIXME: you have won
	    return;
	}
	printf("index %p\n", index);
	// wave [n]
	if (ConfigGetInteger(array, &ival, "next-wave", NULL)) {
	    TdVars->NextWave = ival;
	} else {
	    Warning("no next-wave counter found\n");
	    TdVars->NextWave = 500;
	}

	TdVars->WaveGroup = 0;
	TdVars->GroupInterval = 0;
	TdVars->WaveCounter = -1;
    }
    --TdVars->NextWave;
    //
    //	Next group of units
    //
    if (TdVars->GroupInterval >= 0) {
	if (!TdVars->GroupInterval) {
	    ConfigGetArray(ConfigDict(TdVars->Config), &array, "wave", NULL);
	    // waves
	    index = ConfigNewInteger(TdVars->Wave);
	    array = ConfigArrayFirstFixedKey(array, &index);
	    // waves[n]
	    index = ConfigNewInteger(TdVars->WaveGroup);
	    value = ConfigArrayFirstFixedKey(array, &index);
	    if (!value) {
		TdVars->Wave++;
		TdVars->GroupInterval = -1;
		return;
	    }
	    array = value;
	    // groups[n]

	    // first time get number of units
	    if (TdVars->WaveCounter == -1) {
		TdVars->WaveCounter = 1;
		if (ConfigGetInteger(array, &ival, "count", NULL)) {
		    printf("count %zd\n", ival);
		    TdVars->WaveCounter = ival;
		}
	    }
	    // create count units
	    if (TdVars->WaveCounter--) {
		TdCreateUnit(array);
	    } else {
		TdVars->WaveGroup++;
		TdVars->WaveCounter = -1;
	    }
	    if (ConfigGetInteger(array, &ival, "interval", NULL)) {
		// printf("interval %zd\n", ival);
		TdVars->GroupInterval = ival;
	    }
	}
	--TdVars->GroupInterval;
    }
}

// ------------------------------------------------------------------------ //

/**
**	Cache tower type configuration.
*/
static void TdCacheTowerTypes(void)
{
    const ConfigObject *array;

    //
    //	array of tower types
    //
    if (ConfigGetArray(ConfigDict(TdVars->Config), &array, "tower", NULL)) {
	const ConfigObject *index;
	const ConfigObject *value;
	int i;

	//
	//	towers
	//
	i = 0;
	index = NULL;
	value = ConfigArrayFirstFixedKey(array, &index);
	while (value) {
	    ssize_t ival;

	    if (ConfigGetInteger(value, &ival, "type", NULL)) {
		//TdVars->TowerTypes[i].Type = ival;
	    }
	    if (ConfigGetInteger(value, &ival, "range", NULL)) {
		TdVars->TowerTypes[i].Range = ival;
	    }
	    if (ConfigGetInteger(value, &ival, "damage", NULL)) {
		TdVars->TowerTypes[i].Damage = ival;
	    }
	    if (ConfigGetInteger(value, &ival, "speed", NULL)) {
		TdVars->TowerTypes[i].Speed = ival;
	    }
	    if (ConfigGetInteger(value, &ival, "cost", NULL)) {
		TdVars->TowerTypes[i].Cost = ival;
	    }
	    if (ConfigGetInteger(value, &ival, "build-time", NULL)) {
		TdVars->TowerTypes[i].Time = ival;
	    }
	    value = ConfigArrayNextFixedKey(array, &index);
	    ++i;
	}
    }
}

/**
**	Close and destroy tower defense window.
*/
static void TdDestroy(void)
{
    xcb_destroy_window(Connection, TdVars->Window);
    TdVars->Window = XCB_NONE;

    xcb_free_pixmap(Connection, TdVars->Pixmap);
    xcb_free_pixmap(Connection, TdVars->Working);

    ConfigFreeMem(TdVars->Config);
}

/**
**	Create the tower defense window.
**
**	@param filename start map or game.
*/
void TdCreate(const char *filename)
{
    uint32_t values[3];
    xcb_window_t window;
    xcb_size_hints_t size_hints;
    int x;
    int y;
    int width;
    int height;
    Config *config;
    char *name;
    const char *sval;

    if (TdVars->Window) {		// already running
	TdDestroy();
	return;
    }
    //
    // load config
    //
    name = ExpandPath(filename);
    config = ConfigReadFile(0, NULL, name);
    free(name);

    if (!config) {
	fprintf(stderr, "error parsing config file '%s'\n", filename);
	return;
    }

    if (ConfigGetString(ConfigDict(config), &sval, "health-bar", "color",
	    NULL)) {
	xcb_coloritem_t temp;

	ColorParse(sval, &temp);
	TdVars->HealthBarPixel = temp.pixel;
    }

    TdVars->Config = config;

    x = 0;
    y = 0;
    width =
	XcbScreen->width_in_pixels < 800 ? XcbScreen->width_in_pixels : 800;
    height =
	XcbScreen->height_in_pixels < 480 ? XcbScreen->height_in_pixels : 480;

    // display pixmap
    TdVars->Pixmap = xcb_generate_id(Connection);
    xcb_create_pixmap(Connection, XcbScreen->root_depth, TdVars->Pixmap,
	XcbScreen->root, width, height);
    // working pixmap
    TdVars->Working = xcb_generate_id(Connection);
    xcb_create_pixmap(Connection, XcbScreen->root_depth, TdVars->Working,
	XcbScreen->root, width, height);

    window = xcb_generate_id(Connection);

    values[0] = TdVars->Pixmap;
    values[1] = Colors.PanelBG.Pixel;
    values[2] =
	XCB_EVENT_MASK_BUTTON_PRESS | XCB_EVENT_MASK_BUTTON_RELEASE |
	XCB_EVENT_MASK_POINTER_MOTION | XCB_EVENT_MASK_EXPOSURE;
    xcb_create_window(Connection, XCB_COPY_FROM_PARENT, window,
	XcbScreen->root, x, y, width, height, 0, XCB_WINDOW_CLASS_INPUT_OUTPUT,
	XCB_COPY_FROM_PARENT,
	XCB_CW_BACK_PIXMAP | XCB_CW_BACK_PIXEL | XCB_CW_EVENT_MASK, values);

    size_hints.flags = 0;
    xcb_size_hints_set_position(&size_hints, 0, x, y);
    xcb_set_wm_size_hints(Connection, window, WM_NORMAL_HINTS, &size_hints);
    xcb_set_wm_name(Connection, window, STRING, sizeof("tower-defense") - 1,
	"tower-defense");

    TdVars->Height = height;
    TdVars->Width = width;
    TdVars->Window = window;

    xcb_map_window(Connection, window);

    TdVars->Lives = 15;
    TdVars->Credits = 50;
    TdVars->Points = 0;

    TdVars->Wave = 0;
    TdVars->NextWave = 50;
    TdVars->GroupInterval = -1;

    TdCacheTowerTypes();

    TdVars->UnitN = 0;
    TdPrepareMap();
    TdMoveMap();

    TdDraw(0);
    TdVars->Ticks = GetMsTicks();
}

/**
**	Handle an expose event on a tower defense window.
**
**	@param event	x11 expose event
**
**	@returns true if the event was for the tower defense module,
**	0 otherwise.
*/
int TdHandleExpose(const xcb_expose_event_t * event)
{
    if ( /* TdVars && */ event->window == TdVars->Window) {
	xcb_aux_clear_window(Connection, TdVars->Window);
	return 1;
    }
    return 0;
}

/**
**	Handle a mouse button press event.
**
**	@param event	X11 button press event
**
**	@returns true if event was handled, false othewise.
*/
int TdHandleButtonPress(const xcb_button_press_event_t * event)
{
    int tx;
    int ty;

    if ( /* TdVars && */ event->event == TdVars->Window) {
	if (event->detail == XCB_BUTTON_INDEX_1 && event->event_x < 32
	    && event->event_y < 32) {
	    TdDestroy();
	    return 1;
	}
	TdPixel2Tile(event->event_x, event->event_y, &tx, &ty);
	if (tx < 0 || ty < 0 || tx >= MAP_W || ty >= MAP_H) {
	    printf("out side map\n");
	} else if (TdVars->Tiles[tx + ty * MAP_W].Distance == 9999) {
	    printf("border map\n");
	} else {
	    TdVars->Tiles[tx + ty * MAP_W].Tower =
		!TdVars->Tiles[tx + ty * MAP_W].Tower;
	    TdMoveMap();
	    // FIXME: blocks way?
	}

	return 1;
    }
    return 0;
}

/**
**	Handle a mouse button release event.
**
**	@param event	X11 button release event
**
**	@returns true if event was handled, false othewise.
*/
int TdHandleButtonRelease(const xcb_button_release_event_t * event)
{
    if ( /* !TdVars || */ event->event != TdVars->Window) {
	return 0;
    }
    return 1;
}

/**
**	Handle an motion notify event on tower defense window.
**
**	@param event	X11 motion notify event
**
**	@returns true if the event was for the tower defense module,
**	0 otherwise.
*/
int TdHandleMotionNotify(const xcb_motion_notify_event_t * event)
{
    if ( /* !TdVars || */ event->event != TdVars->Window) {
	return 0;
    }
    event = event;
    //
    return 1;
}

/**
**	Timeout for tower defense.  Max resolution is * 50ms (20Hz)
**
**	@param tick	current tick in ms
**	@param x	current mouse x-coordinate
**	@param y	current mouse y-coordinate
*/
void TdTimeout(uint32_t tick, int __attribute__ ((unused)) x, int
    __attribute__ ((unused)) y)
{
    uint32_t st;
    uint32_t et;

    if ( /* !TdVars || */ !TdVars->Window) {
	return;
    }
    if (tick < TdVars->Ticks + 50) {
	return;
    }
    TdVars->Ticks += 50;

    st = GetMsTicks();
    TdUpdateWave();
    TdMoveUnits();
    TdUpdateTowers();
    TdDraw(0);
    et = GetMsTicks();

    TdVars->FrameTime = abs(et - st);
    //xcb_aux_sync(Connection);
}

#endif // } USE_TD

/// @}
